"""Anything used to construct a websocket endpoint"""
import asyncio
import json
import logging
from typing import Any
from urllib.parse import parse_qsl

from channels.generic.websocket import AsyncJsonWebsocketConsumer
from idom.core.dispatcher import dispatch_single_view
from idom.core.layout import Layout, LayoutEvent

from .config import IDOM_REGISTERED_COMPONENTS


_logger = logging.getLogger(__name__)


class IdomAsyncWebSocketConsumer(AsyncJsonWebsocketConsumer):
    """Communicates with the browser to perform actions on-demand."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    async def connect(self) -> None:
        await super().connect()
        self._idom_dispatcher_future = asyncio.ensure_future(self._run_dispatch_loop())

    async def disconnect(self, code: int) -> None:
        if self._idom_dispatcher_future.done():
            await self._idom_dispatcher_future
        else:
            self._idom_dispatcher_future.cancel()
        await super().disconnect(code)

    async def receive_json(self, content: Any, **kwargs: Any) -> None:
        await self._idom_recv_queue.put(LayoutEvent(**content))

    async def _run_dispatch_loop(self):
        view_id = self.scope["url_route"]["kwargs"]["view_id"]

        try:
            component_constructor = IDOM_REGISTERED_COMPONENTS[view_id]
        except KeyError:
            _logger.warning(f"Uknown IDOM view ID {view_id!r}")
            return

        query_dict = dict(parse_qsl(self.scope["query_string"].decode()))
        component_kwargs = json.loads(query_dict.get("kwargs", "{}"))

        try:
            component_instance = component_constructor(**component_kwargs)
        except Exception:
            _logger.exception(
                f"Failed to construct component {component_constructor} "
                f"with parameters {component_kwargs}"
            )
            return

        self._idom_recv_queue = recv_queue = asyncio.Queue()
        try:
            await dispatch_single_view(
                Layout(component_instance),
                self.send_json,
                recv_queue.get,
            )
        except Exception:
            await self.close()
            raise



import asyncio
import functools
import os

from django.core.cache import caches
from django.http import HttpRequest, HttpResponse
from idom.config import IDOM_WED_MODULES_DIR

from .config import IDOM_WEB_MODULE_CACHE, IDOM_WEB_MODULE_LRU_CACHE_SIZE


if IDOM_WEB_MODULE_CACHE is None:

    def async_lru_cache(*lru_cache_args, **lru_cache_kwargs):
        def async_lru_cache_decorator(async_function):
            @functools.lru_cache(*lru_cache_args, **lru_cache_kwargs)
            def cached_async_function(*args, **kwargs):
                coroutine = async_function(*args, **kwargs)
                return asyncio.ensure_future(coroutine)

            return cached_async_function

        return async_lru_cache_decorator

    @async_lru_cache(IDOM_WEB_MODULE_LRU_CACHE_SIZE)
    async def web_modules_file(request: HttpRequest, file: str) -> HttpResponse:
        file_path = IDOM_WED_MODULES_DIR.current.joinpath(*file.split("/"))
        return HttpResponse(file_path.read_text(), content_type="text/javascript")


else:
    _web_module_cache = caches[IDOM_WEB_MODULE_CACHE]

    async def web_modules_file(request: HttpRequest, file: str) -> HttpResponse:
        file = IDOM_WED_MODULES_DIR.current.joinpath(*file.split("/")).absolute()
        last_modified_time = os.stat(file).st_mtime
        cache_key = f"{file}:{last_modified_time}"

        response = _web_module_cache.get(cache_key)
        if response is None:
            response = HttpResponse(file.read_text(), content_type="text/javascript")
            _web_module_cache.set(cache_key, response, timeout=None)

        return response



import json
import sys
from importlib import import_module
from urllib.parse import urlencode
from uuid import uuid4

from django import template

from django_idom.config import (
    IDOM_REGISTERED_COMPONENTS,
    IDOM_WEB_MODULES_URL,
    IDOM_WEBSOCKET_URL,
)


register = template.Library()


@register.inclusion_tag("idom/view.html")
def idom_component(_component_id_, **kwargs):
    _register_component(_component_id_)

    json_kwargs = json.dumps(kwargs, separators=(",", ":"))

    return {
        "idom_websocket_url": IDOM_WEBSOCKET_URL,
        "idom_web_modules_url": IDOM_WEB_MODULES_URL,
        "idom_mount_uuid": uuid4().hex,
        "idom_component_id": _component_id_,
        "idom_component_params": urlencode({"kwargs": json_kwargs}),
    }


def _register_component(full_component_name: str) -> None:
    module_name, component_name = full_component_name.rsplit(".", 1)

    if module_name in sys.modules:
        module = sys.modules[module_name]
    else:
        try:
            module = import_module(module_name)
        except ImportError as error:
            raise RuntimeError(
                f"Failed to import {module_name!r} while loading {component_name!r}"
            ) from error

    try:
        component = getattr(module, component_name)
    except AttributeError as error:
        raise RuntimeError(
            f"Module {module_name!r} has no component named {component_name!r}"
        ) from error

    IDOM_REGISTERED_COMPONENTS[full_component_name] = component



from django.urls import path

from . import views
from .config import IDOM_WEB_MODULES_URL, IDOM_WEBSOCKET_URL
from .websocket_consumer import IdomAsyncWebSocketConsumer


IDOM_WEBSOCKET_PATH = path(
    IDOM_WEBSOCKET_URL + "<view_id>/", IdomAsyncWebSocketConsumer.as_asgi()
)
"""A URL resolver for :class:`IdomAsyncWebSocketConsumer`

While this is relatively uncommon in most Django apps, because the URL of the
websocket must be defined by the setting ``IDOM_WEBSOCKET_URL``. There's no need
to allow users to configure the URL themselves.
"""


IDOM_WEB_MODULES_PATH = path(
    IDOM_WEB_MODULES_URL + "<path:file>", views.web_modules_file
)
"""A URL resolver for static web modules required by IDOM

While this is relatively uncommon in most Django apps, because the URL of the
websocket must be defined by the setting ``IDOM_WEBSOCKET_URL``. There's no need
to allow users to configure the URL themselves.
"""



from typing import Dict

from django.conf import settings
from django.core.cache import DEFAULT_CACHE_ALIAS
from idom.core.proto import ComponentConstructor


IDOM_REGISTERED_COMPONENTS: Dict[str, ComponentConstructor] = {}

IDOM_BASE_URL = getattr(settings, "IDOM_BASE_URL", "_idom/")
IDOM_WEBSOCKET_URL = IDOM_BASE_URL + "websocket/"
IDOM_WEB_MODULES_URL = IDOM_BASE_URL + "web_module/"

_CACHES = getattr(settings, "CACHES", {})
if _CACHES:
    if "idom_web_modules" in getattr(settings, "CACHES", {}):
        IDOM_WEB_MODULE_CACHE = "idom_web_modules"
    else:
        IDOM_WEB_MODULE_CACHE = DEFAULT_CACHE_ALIAS
else:
    IDOM_WEB_MODULE_CACHE = None


# the LRU cache size for the route serving IDOM_WEB_MODULES_DIR files
IDOM_WEB_MODULE_LRU_CACHE_SIZE = getattr(
    settings, "IDOM_WEB_MODULE_LRU_CACHE_SIZE", None
)
